/*
 * A flexpage abstraction for handling map requests.
 *
 * Copyright (C) 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/consts.h>
#include <l4/re/c/dataspace.h>
#include <l4/sys/task.h>

#include <stdio.h>
#include <string.h>

#include "flexpage.h"



/* Round down the given value to the largest possible page size. */

static unsigned long trunc_multiple(unsigned long value)
{
  unsigned char order = L4_PAGESHIFT;
  unsigned char last = order;
  unsigned long multiple = 1UL << order;

  while (order < 30)
  {
    if (value < multiple)
      break;

    last = order;
    order += 1;
    multiple <<= 1;
  }

  return l4_trunc_size(value, last);
}



/* A flexpage abstraction for a map request. */

void Flexpage::reset(unsigned long offset, l4_addr_t hot_spot, int flags)
{
  _flags = flags;

  /* The hot spot is already page aligned and is used to determine the flexpage
     dimensions. Since it cannot identify a page beyond the boundaries of
     available memory, the size of this memory is used to constrain it. */

  _page_addr = _buffer + (hot_spot % trunc_multiple(_buffer_size));

  /* A suitable size given the accessed address.

     It seems that the l4_fpage_max_order function will only combine the hot
     spot with the supplied address in the anticipated way if the address is
     already aligned to a page. */

  _order = l4_fpage_max_order(L4_PAGESHIFT, _page_addr, _buffer,
                              _buffer + _buffer_size, hot_spot);

  /* Get the supposed base of the flexpage. */

  _base_addr = l4_trunc_size(_page_addr, _order);

  /* Get the dataspace offset for the base of the flexpage. */

  _base_offset = l4_trunc_page(offset - (_page_addr - _base_addr));

  /* Prepare the item. */

  upgrade(flags);
}

/* Return a flexpage item with the specified flags. */

l4_fpage_t Flexpage::get_fpage(unsigned long flags)
{
  return l4_fpage(_base_addr, _order,
                  (flags & L4RE_DS_MAP_FLAG_RW) ? L4_FPAGE_RW
                                                : L4_FPAGE_RO);
}

/* Return whether the flexpage overlaps the given data limits. */

bool Flexpage::overlap(unsigned long start, unsigned long end)
{
  unsigned long data_end = _data_start + get_size();

  return (_data_start < end) && (data_end > start);
}

/* Convenience method for padding a flexpage. */

void Flexpage::pad()
{
  memset((void *) (_base_addr + _data_length), 0, get_size() - _data_length);
}

void Flexpage::clear()
{
  memset((void *) _base_addr, 0, get_size());
}

/* Associate the flexpage with a particular region of data starting at the given
   position and constrained by the total size of the data. */

void Flexpage::set_data_region(unsigned long data_position,
                               unsigned long data_size)
{
  /* Obtain the page offset and length of the data in the flexpage. */

  _data_start = data_position + _base_offset;
  set_data_size(data_size);
}

void Flexpage::set_data_size(unsigned long data_size)
{
  unsigned long data_end = _data_start + (1 << _order);

  /* The end of the data must not be beyond the end of the file. */

  if (data_end > data_size)
    _data_length = data_size - _data_start;
  else
    _data_length = data_end - _data_start;
}

/* Return whether this flexpage supports the indicated position. */

bool Flexpage::supports(unsigned long data_position)
{
  unsigned long end = _data_start + get_size();
  unsigned long target = data_position;

  return (_data_start <= target) && (target < end);
}

bool Flexpage::supports_flags(unsigned long flags)
{
  return (flags == _flags) || (!flags);
}

/* Upgrade the flexpage. */

void Flexpage::upgrade(unsigned long flags)
{
  _flags = flags;
}

/* Revoke the flexpage's mappings. */

void Flexpage::unmap()
{
  l4_task_unmap(L4RE_THIS_TASK_CAP, get_fpage(_flags), L4_FP_OTHER_SPACES);
}

/* Return whether the flexpage is writable. */

int Flexpage::writable()
{
  return _flags & L4RE_DS_MAP_FLAG_RW;
}
