/*
 * A pager abstraction for handling map requests.
 *
 * Copyright (C) 2019, 2020 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/re/consts.h>

#include <stdio.h>

#include "dataspace_server.h"

#include "flexpage.h"
#include "pager.h"



/* Debugging helpers. */

static void show_flexpage(Flexpage *flexpage, const char *label)
{
  printf("%s: %p: %lx + %lx for %ld + %ld\n", label, flexpage,
                                              flexpage->get_base(),
                                              flexpage->get_size(),
                                              flexpage->get_data_start(),
                                              flexpage->get_data_length());
}



/* Handle a map request by allocating and associating a flexpage with the
   requested memory region, populating the flexpage with data. */

long Pager::map(unsigned long offset, l4_addr_t hot_spot, unsigned long flags,
                l4_snd_fpage_t *region)
{
  printf("Pager::map(offset=%lx, hot_spot=%lx, flags=%lx)\n",
    offset, hot_spot, flags);

  /* Find any flexpage supporting the request. */

  Flexpage *flexpage = get_accessor()->get_flexpage(get_data_start(),
                                                    get_data_size(),
                                                    offset, hot_spot, flags);

  show_flexpage(flexpage, "new flexpage");

  /* Send the flexpage explicitly. */

  region->fpage = flexpage->get_fpage(flags);
  region->snd_base = hot_spot;

  long err = complete_Dataspace_map(*region);

  if (err)
    return err;

  /* Make the sent flexpage available for reuse. */

  get_accessor()->queue_flexpage(flexpage);

  return IPC_MESSAGE_SENT;
}



/* Flush all allocated flexpages within the pager's region, writing data to
   storage. If the keep indicator is set to a true value, the flexpages will be
   retained; otherwise, they are invalidated. */

void Pager::flush_all(bool keep)
{
  unsigned long start = get_data_start();
  unsigned long end = start + get_data_size();

  get_accessor()->flush_all(start, end, keep);
}
